package com.bergerkiller.bukkit.nolagg;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.bukkit.Bukkit;
import org.bukkit.block.BlockFace;
import org.bukkit.entity.Player;

import org.bukkit.craftbukkit.CraftServer;
import org.bukkit.craftbukkit.entity.CraftPlayer;

import net.minecraft.server.Chunk;
import net.minecraft.server.ChunkCoordIntPair;
import net.minecraft.server.EntityPlayer;
import net.minecraft.server.Packet53BlockChange;
import net.minecraft.server.World;

@SuppressWarnings({"rawtypes", "unchecked"})
public class ChunkSendQueue extends LinkedList {

	private static final long serialVersionUID = 1L;
	private static int task = -1;
	public static void init() {
		if (enabled) {
			task = Bukkit.getServer().getScheduler().scheduleSyncRepeatingTask(NoLagg.plugin, new Runnable() {
				public void run() {
					//modify the weight based on global conditions and set the interval/rate
					global(new Operation() {
						public void handle(EntityPlayer ep) {
							ChunkSendQueue queue = bind(ep);
							double chunkspertick = queue.getPercentage(); //1 = high queue 0 = empty queue
							chunkspertick = minRate * (1 - chunkspertick) + maxRate * chunkspertick;
							queue.interval = (int) (1 / chunkspertick);
							if (queue.interval == 0) queue.interval = 1;
							queue.rate = (int) chunkspertick;
							if (queue.rate == 0) queue.rate = 1;
							if (queue.intervalcounter == -1) queue.intervalcounter = queue.interval;
							queue.update();
						}
					});
				}
			}, 1, 1);
		}
	}
	public static void deinit() {
		if (task != -1) {
			Bukkit.getServer().getScheduler().cancelTask(task);
		}
	}
	
	public static void deQueueGlobal(final Chunk chunk) {
		deQueueGlobal(chunk.world, chunk.x, chunk.z);
	}
	public static void queueGlobal(final Chunk chunk) {
		queueGlobal(chunk.world, chunk.x, chunk.z);
	}
	public static void deQueueGlobal(final World world, final int x, final int z) {
		deQueueGlobal(world, new ChunkCoordIntPair(x, z));
	}
	public static void queueGlobal(final World world, final int x, final int z) {
		queueGlobal(world, new ChunkCoordIntPair(x, z));
	}
	public static void deQueueGlobal(final World world, final ChunkCoordIntPair pair) {
		world(world, new Operation() {
			public void handle(EntityPlayer ep) {
				ep.chunkCoordIntPairQueue.remove(pair);
			}
		});
	}
	public static void queueGlobal(final World world, final ChunkCoordIntPair pair) {
		world(world, new Operation() {
			public void handle(EntityPlayer ep) {
				ep.chunkCoordIntPairQueue.add(0, pair);
			}
		});
	}
	
	private static <T extends Operation> T global(T op) {
		op.init();
		for (World world : ((CraftServer) Bukkit.getServer()).getServer().worlds) {
			world(world, op);
		}
		return op;
	}
	private static <T extends Operation> T world(World world, T op) {
		op.init(world);
		for (Object o : world.players) op.handle((EntityPlayer) o);
		return op;
	}
	private static class Operation {
		public void handle(EntityPlayer ep) {}
		public void init(World world) {}
		public void init() {}
	}
	
	public static boolean enabled = true;
	public static ChunkSendQueue bind(Player with) {
		if (!enabled) return null;
		return bind(((CraftPlayer) with).getHandle());
	}
	public static ChunkSendQueue bind(EntityPlayer with) {
		if (!enabled) return null;
		if (!(with.chunkCoordIntPairQueue instanceof ChunkSendQueue)) {
			with.chunkCoordIntPairQueue = new ChunkSendQueue(with);
		}
		return (ChunkSendQueue) with.chunkCoordIntPairQueue;
	}	
	private ChunkSendQueue(EntityPlayer ep) {
		this.ep = ep;
		this.update();
	}
	
	public static int initialBoost = 0; //the size that has to be downloaded
	public static double maxRate = 2;
	public static double minRate = 0.25;
		
	public void boost(int size) {
		this.boost += size;
	}
	
	private EntityPlayer ep;
	private BlockFace sendDirection = BlockFace.NORTH;
	private int x;
	private int z;
	private int interval = 4;
	private int rate = 1;
	private int boost = 25;
	private int intervalcounter = -1;
	
	public double getPercentage() {
		return getPercentage(this.ep);
	}
	public static double getPercentage(EntityPlayer ep) {
		return (double) ep.chunkCoordIntPairQueue.size() / (double) Util.chunkarea;
	}
			
	private List<ChunkCoordIntPair> toTrigger = new ArrayList<ChunkCoordIntPair>();
	
	private void update() {
		if (this.intervalcounter >= this.interval)  {
			//sorting
			BlockFace newDirection = Util.getDirection((int) this.ep.yaw - 90);
			int newx = (int) (ep.locX + ep.motX * 10) >> 4;
			int newz = (int) (ep.locZ + ep.motZ * 10) >> 4;
			if (newx != this.x || newz != this.z || this.sendDirection != newDirection) {
				this.sendDirection = newDirection;
				this.x = newx;
				this.z = newz;
				//sort self
				if (!this.isEmpty()) {
					java.util.Collections.sort(this, new ChunkCoordComparator(this.x, this.z, this.sendDirection));
					//download speed boost when the chunk below the player has to be sent
					ChunkCoordIntPair first = (ChunkCoordIntPair) super.get(0);
					if (first.x == this.x && first.z == this.z) this.boost(initialBoost);
				}
			}
			
			//send triggers
			synchronized (this.toTrigger) {
				for (ChunkCoordIntPair pair : this.toTrigger) {
					//send a single block packet for this chunk
					this.ep.netServerHandler.sendPacket(new Packet53BlockChange(pair.x * 16, 0, pair.z * 16, this.ep.world));
				}
				this.toTrigger.clear();
			}
			
			//send using rate
			int counter = 0;
			while (!this.isEmpty() && (counter < this.rate || this.hasBoost())) {
				ChunkCoordIntPair pair = (ChunkCoordIntPair) super.remove(0);
				ChunkHandler.sendChunk(this.ep.world.getChunkAt(pair.x, pair.z), ep);
				synchronized(this.toTrigger) {
					this.toTrigger.add(pair);
				}
				counter++;
			}
			this.intervalcounter = 1;
		} else {
			this.intervalcounter++;
		}
	}
	
	private boolean hasBoost() {
		if (this.boost > 0) {
			--this.boost;
			return true;
		}
		return false;
	}
		
	/*
	 * get(0) is called in EntityPlayer to get the next chunk to send
	 */
	public Object get(int index) {
		return null;
	}
	
	/*
	 * remove(o) is called in EntityPlayer to remove a sent chunk coordinate
	 * remove(o) is called in PlayerInstance to remove a chunk out of reach
	 */
	
	/*
	 * add(0) is called in PlayerInstance to queue a new chunk coordinate
	 */
	public boolean add(Object object) {
		this.sendDirection = null; //invalidate
		return super.add(object);
	}

}
